#!/usr/bin/env python
#! -*- coding: utf-8 -*-

from gi.repository import Clutter, Cogl, GdkPixbuf
import videoplayer
import os, re, sys

vid = None
resources = os.path.join(os.path.abspath(sys.path[0]), 'resources')

class Item:
    # Cogl crashes if I try to load a texture > MAX_SIZE
    MAX_SIZE = 1600
    PREVIEW_HEIGHT = 100
    def __init__(self, presentation, attrs):
	# Default attribute values
        vals = { 'id' : 0, 'parent-id': 0, 'rotation': 0,
                 'home': None, 'home-size' : None,
		 'background-colour': (0,0,0,0) }
        vals.update(attrs)
	self.presentation = presentation
	try:
	    self.max_size = min(self.MAX_SIZE, \
 	        max(presentation.slide_size[0], presentation.slide_size[1]))
	except:
	    self.max_size = self.MAX_SIZE

        self.iid = vals['id']
	self.parent_iid = vals['parent-id']
        self.rotation = vals['rotation']
        self.home = vals['home']
        self.home_size = vals['home-size']
        self.scale = (1.0, 1.0)
        self.expanded = False
	self.editing = False
	self.slide = None
	self.background_colour = vals['background-colour']

        self.drag = Clutter.DragAction.new()
        self.drag.connect('drag-begin', self.on_drag_begin)
        self.drag.connect('drag-end', self.on_drag_end)

	actor = self.actor = Clutter.Actor.new()
	self.content_actor = Clutter.Actor.new()
        self.actor.add_child(self.content_actor)
	c = Clutter.Color.new(*self.background_colour)
	self.content_actor.set_background_color(c)

        self.content_actor.set_pivot_point(0.5, 0.5)
        self.content_actor.set_content_gravity(Clutter.ContentGravity.RESIZE_ASPECT)
        self.rotate()
    
        # Connect signal handlers for events
        actor.set_reactive(True)
        actor.connect('button-press-event', self.on_button_press)
        actor.connect('key-press-event', self.on_key_press)
        actor.connect('transitions-completed', self.on_transitions_done)
        actor.connect('key-focus-in', self.on_key_focus)
        actor.connect('key-focus-out', self.on_lost_key_focus)
    def dump_assets(self):
	return

    def start_edit(self):
	if self.editing:
	    return
	self.editing = True
	#print "Start edit", self.iid
        self.actor.add_action(self.drag)
    def stop_edit(self):
	if not self.editing:
	    return
	self.editing = False 
        self.actor.remove_action(self.drag)
    def on_key_focus(self, actor):
	return False
    def on_lost_key_focus(self, actor):
	return False
    def show(self):
	self.actor.show_all()
        if self.home is not None:
            self.set_position(*self.home)
	if self.home_size is not None:
	    self.set_size(*self.home_size)
    def reparent(self, slide, actor):
        self.slide = slide
	actor.add_child(self.actor)
    def unparent(self):
	if self.slide is None:
	   return
	# FIXME: Remove from parent slide and relocate to stage

    def set_default_size(self, actor):
        pixbuf_height = self.image_size[1]
        scale = self.PREVIEW_HEIGHT / float(pixbuf_height) if pixbuf_height else 0
        w, h = self.get_visible_size()
 	w *= scale
	h *= scale
        self.set_size(w, h)
	#print "Set default size %d, %d for image %d, %d" % \
	    #(w, h, self.image_size[0], self.image_size[1])

    def get_visible_size(self):
        w,h = self.actor.get_size()
	x,y = self.scale
        return (w * x, h * y)
    def get_size(self):
        return self.home_size

    def set_position(self, x, y):       
        self.home = (x, y)
        self.actor.set_position(x, y)
        #print "Item %d moved to %d, %d rotation %d size %d, %d" % \
	    #(self.iid, x, y, self.rotation,
             #self.get_visible_size()[0], self.get_visible_size()[1])
    def has_position(self):
        return self.home is not None
    def get_position(self):
        return self.actor.get_position()

    def calc_scale_from_size(self, w, h):
	nat_w, nat_h = self.actor.get_size()
        scale = (w / nat_w, h / nat_h)
	return scale

    def set_size(self, w, h):
	self.home_size = (w, h)
	self.actor.set_size(w, h)
        self.content_actor.set_size(w, h)
	# print "Size of %d now" % (self.iid), self.home_size
	# self.set_scaled_size(w, h)

    def set_scaled_size(self, w, h):
	self.scale = self.calc_scale_from_size(w, h)
        self.actor.set_scale(*self.scale)
	# print "Scale of %d now" % (self.iid), self.scale

    def on_transitions_done(self, actor):
        pass

    def on_button_press(self, actor, event):
        # print "Got button press for item", self.iid
        self.actor.grab_key_focus()

        if event.click_count > 1:
            self.slide.expand_item(self)
            return True
        return False

    def on_key_press(self, stage, event):
        code = event.keyval
        # print "Got key press for item", self.iid
        if code == Clutter.KEY_r:
           self.rotate(90)
           return True
        elif code == Clutter.KEY_R:
           self.rotate(-90)
           return True
        elif code == Clutter.KEY_plus:
           self.embiggen()
           return True
        elif code == Clutter.KEY_minus:
	   self.emshrinken()
           return True
	elif code == Clutter.KEY_Escape:
	   if self.expanded:
		self.unexpand()
	   	return True
        return False

    def on_drag_begin(self, drag, actor, x, y, modifiers):
        # print "Drag began at (%d, %d)" % self.home
        return False

    def on_drag_end(self, drag, actor, x, y, modifiers):
        self.home = self.get_position()
        # print "Drag ended at (%d, %d)" % self.home
        return False

    def rotate(self, angle=0):
        self.rotation += angle
	# print "Rotation for %d now %f" % (self.iid, self.rotation)
        self.content_actor.set_rotation_angle(Clutter.RotateAxis.Z_AXIS, self.rotation)
    def embiggen(self):
        w, h = self.get_size()
	self.set_size(w + 10, h * (w + 10)/w)

    def emshrinken(self):
        w, h = self.get_size()
        new_w = w - 10 if w - 10 > 0 else w
        self.set_size(new_w, h * (new_w)/w)

    def serialise(self):
        return { 'id' : self.iid, 'parent-id' : self.parent_iid,
                 'home': self.home,
		 'home-size' : self.home_size,
                 'rotation' : self.rotation,
		 'background-colour' : self.background_colour }

    def expand(self, x, y, w, h):
	# print "Expand %d to %g, %g, %g, %g" % (self.iid, x, y, w, h)
	actor = self.actor
        actor.set_easing_duration(200)
        actor.set_easing_mode(Clutter.AnimationMode.LINEAR)
        actor.set_position(x, y)
        self.set_scaled_size(w, h)
        self.expanded = True

    def unexpand(self):
        # print "Unexpanding", self.iid, self
	actor = self.actor
        actor.set_easing_duration(100)
        actor.set_easing_mode(Clutter.AnimationMode.LINEAR)
        self.set_scaled_size(*self.home_size)
        actor.set_position(*self.home)
        # print "unexpanded size %d, %d @ %d, %d" % (self.home_size[0], self.home_size[1], self.home[0], self.home[0])
        self.expanded = False

    def hide(self):
	self.actor.hide()
    def set_opacity(self, opacity):
	self.actor.set_opacity(opacity)

class ImageItem(Item):
    def __init__(self, presentation, attrs):
        Item.__init__(self, presentation, attrs)

        # Check for prescaled version and load that in preference
        path = attrs['path']
	self.orig_path = path
        self.scaled_path = path + '_scaled.png'
	self.path = path
	self.loaded = False
        if os.path.exists(self.scaled_path):
            self.path = self.scaled_path
    def dump_assets(self):
	print self.path

    def serialise(self):
        i = Item.serialise(self)
        i.update({ 'type' : 'ImageItem', 'path' : self.orig_path })
        return i

    def limitSize(self, pixbuf):
        w = pixbuf.get_width()
        h = pixbuf.get_height()
        scaled = False
        if w > self.max_size:
            w = self.max_size
            h = (w * h) / pixbuf.get_width()
            scaled = True
        if h > self.max_size:
            w = (w * self.max_size) / h
            h = self.max_size
            scaled = True
        p = pixbuf.scale_simple(w, h, GdkPixbuf.InterpType.BILINEAR)
        if scaled:
            # Save out scaled version
            p.savev(self.scaled_path, 'png', [], [])
        return p
                
    def load_image(self):
	if self.loaded:
	    return

        pixbuf = GdkPixbuf.Pixbuf.new_from_file(self.path)
        #print "loading file %s, size %d,%d" % \
        #    (self.path, pixbuf.get_width(), pixbuf.get_height())
        pixbuf = self.limitSize(pixbuf)
        image = Clutter.Image.new();
        if pixbuf.get_has_alpha():
            fmt = Cogl.PixelFormat.RGBA_8888
        else:
            fmt = Cogl.PixelFormat.RGB_888
        w, h = pixbuf.get_width(), pixbuf.get_height()
	self.image_size=(w,h)

        if not image.set_data(pixbuf.get_pixels(), fmt,
               w, h, pixbuf.get_rowstride()):
            print "Failed to load image into texture"

        actor = self.actor
        self.content_actor.set_content(image)
        self.content_actor.set_size(w, h)

        # Make sure that all images are shown with the same height
	if self.home_size is None:
           self.set_default_size(self.actor)
	else:
	   self.set_size(*self.home_size)
    
	self.loaded = True
    def show(self):
	self.load_image()
	Item.show(self)

def ImageItem_deserialise(pres, d):
        return ImageItem(pres, d)

class TextItem(Item):
    def __init__(self, presentation, attrs):
        vals = { 'text' : 'Edit Me',
		 'font' : 'Sans',
                 'font-size' : 10,
                 'background-colour' : (0, 0, 0, 80),
                 'colour' : (255, 255, 255, 255)
               }
	vals.update(attrs)

        Item.__init__(self, presentation, vals)

	if 'path' in vals.keys():
	   text = open(vals['path'], 'r').read()
	else:
	   text = vals['text']

	self.font = vals['font']
	self.font_size = vals['font-size']
        self.colour = vals['colour']
	self.text = text
	self.text_editing = False

        # Load the text into a text actor
        self.content = Clutter.Text.new()
	self.content.set_margin_top(2)
	self.content.set_margin_bottom(2)
	self.content.set_margin_left(5)
	self.content.set_margin_right(5)
        self.actor.add_child(self.content)

        self.content.set_markup(self.text)
	if self.content.get_text() == '':
	  self.content.set_use_markup(False)
	  self.content.set_text(self.text)

	self.set_font(self.font, self.font_size)
	c = Clutter.Color.new(*self.colour)
	self.content.set_color(c)

	if self.home_size is None:
           self.home_size = (128, 128)

	self.set_size(*self.content.get_size())
    def set_font(self, font, font_size):
	self.content.set_font_name("%s %d" % (font, font_size))
	self.font = font
	self.font_size = font_size
    def embiggen(self):
	self.set_font(self.font, self.font_size + 1)
	self.set_size(*self.content.get_size())
    def emshrinken(self):
	if self.font_size > 1:
	  self.set_font(self.font, self.font_size - 1)
	self.set_size(*self.content.get_size())

    def serialise(self):
        i = Item.serialise(self)
        i.update({ 'type' : 'TextItem',
                   'font' : self.font, 'font-size' : self.font_size,
		   'colour' : self.colour,
                   'text' : self.text })
        return i
    def stop_edit(self):
	if self.text_editing:
	    self.text = self.content.get_text()
            self.content.set_markup(self.text)
	    if self.content.get_text() == '':
	      self.content.set_use_markup(False)
	      self.content.set_text(self.text)
	    self.text_editing = False
	    self.content.set_editable(False)
	Item.stop_edit(self)

    def on_button_press(self, actor, event):
        self.content.grab_key_focus()
        if event.click_count > 1 and self.editing:
	    self.text_editing = True
	    self.content.set_use_markup(False)
	    self.content.set_text(self.text)
	    self.content.set_editable(True)
            return True
        return False

def TextItem_deserialise(pres, d):
        return TextItem(pres, d)

class VideoItem(ImageItem):
    def __init__(self, presentation, attrs):
	vals = { 'uri' : None }
	vals.update(attrs)

        self.movie_uri = vals['uri']
        self.movie_path = vals['path']

        self.find_thumb()
        attrs['path'] = self.thumb_name

        ImageItem.__init__(self, presentation, attrs)
    def find_thumb(self):
        # Create or load a real thumbnail
        d = os.path.dirname(self.movie_path)
        f = os.path.basename(self.movie_path).split(".")
        f[-1] = "png"
        f[-2] += '_thumb'
        self.thumb_name = os.path.join(d, ".".join(f))
        if not os.path.exists(self.thumb_name):
	    # No png, try jpg
            f[-1] = "jpg"
            self.thumb_name = os.path.join(d, ".".join(f))

        if not os.path.exists(self.thumb_name):
            self.thumb_name = os.path.join(resources, 'fallback_thumb.png')
    def serialise(self):
        i = Item.serialise(self)
        i.update({ 'type' : 'VideoItem', 'path' : self.movie_path,
                   'uri' : self.movie_uri })
        return i
    def dump_assets(self):
	print self.movie_path
	print self.thumb_name
    def expand(self, x, y, w, h):
        global vid
        # Fade out thumbnail
        self.actor.set_opacity(0)
        # Grab the video player and add it
        if vid is None:
            vid = videoplayer.VideoPlayer()
            vid.set_easing_duration(0)
            vid.hide()
            # vid.set_opacity(0)
            vid.set_easing_duration(200)

	if self.movie_uri:
           vid.load(self.movie_uri)
	else:
           vid.load(self.movie_path)
        # Need this, to get the size:
        vid.reparent(self.slide.content)
        vid.preroll()
        vid.set_scaled_size(w, h)
	vw, vh = vid.get_visible_size()
	x += max(0, w - vw)/2
	y += max(0, h - vh)/2
        vid.set_position(x, y)
        vid.show()
        vid.set_opacity(255)
        self.slide.raise_actor(vid.actor)
        Item.expand(self, x, y, w, h)
    def unexpand(self):
        global vid
        Item.unexpand(self)
        # Fade out the video player
        vid.set_opacity(0)
        self.actor.set_opacity(255)
    def on_transitions_done(self, actor):
        # print "Finished transitions on %d" % (self.iid)
        if not self.expanded:
                # Give up the video player
                vid.hide()
        else:
                vid.show()
                vid.play()

def VideoItem_deserialise(pres, d):
        return VideoItem(pres, d)

class SlideItem(Item):
    def __init__(self, presentation, attrs):
        Item.__init__(self, presentation, attrs)

def SlideItem_deserialise(pres, d):
        return SlideItem(pres, d)

class ItemFactory:
    def __init__(self, presentation):
        self.typefuncs = { \
	    'ImageItem' : ImageItem_deserialise,
	    'TextItem' : TextItem_deserialise,
	    'VideoItem' : VideoItem_deserialise,
	    'SlideItem' : SlideItem_deserialise,
	}
        self.presentation = presentation
        self.next_id = presentation.get_max_id() + 1
    def nextIid(self):
        iid = self.next_id
        self.next_id+=1
	return iid

    def loadFile(self, fname):
	print "trying to load", fname
        pres = self.presentation
        if re.search("(?i)(jpg|png)$", fname):
           # Try to load the file as an image
           return ImageItem(pres, {'id' : self.nextIid(), 'path' : fname})
        if re.search("(?i)(mp4|ogg|webm)$", fname):
           # Load the file as a movie
           return VideoItem(pres, {'id' : self.nextIid(), 'path' : fname})

        if re.search("(?i)(txt)$", fname):
           return TextItem(pres, {'id' : self.nextIid(), 'path' : fname})
        return None
    def loadJSON(self, d):
        pres = self.presentation
        i = self.typefuncs[d['type']](pres, d)
        self.next_id = max(self.next_id, i.iid + 1)
        return i
    def newTextItem(self):
        pres = self.presentation
	return TextItem(pres, {'id' : self.nextIid(), 'text' : "Edit Me" })

